--
-- FS19 - Real Wood Harvester
-- @author:    	kenny456 (kenny456@seznam.cz)
-- @history:	v1.0.0.0 - 2021-01-05 - first release
--
RealWoodHarvester = {}
RealWoodHarvester.confDir = getUserProfileAppPath().. "modsSettings/RealWoodHarvester/"
RealWoodHarvester.modDirectory = g_currentModDirectory
RealWoodHarvester.vehicles = {}

RealWoodHarvester.delimbAccuracies = { 70, 80, 90, 100 }
RealWoodHarvester.woodChipsAmounts = { 0, 1, 2, 3, 4, 5 }
RealWoodHarvester.booleans = { true, false }

RealWoodHarvester.delimbAccuraciesTexts = { '70%', '80%', '90%', '100%' }
RealWoodHarvester.woodChipsAmountsTexts = { 'Off', '1 x', '2 x', '3 x', '4 x', '5 x' }
RealWoodHarvester.booleansTexts = { 'On', 'Off' }

RealWoodHarvester.delimbAccuracy = 100
RealWoodHarvester.woodChipsAmount = 1
RealWoodHarvester.showHelp = true

RealWoodHarvester.delimbAccuracyIndex = 4
RealWoodHarvester.woodChipsAmountIndex = 2
RealWoodHarvester.showHelpIndex = 1

function getIndexFromElement(table, element)
	for index, el in pairs(table) do
		if type(element) == 'number' then
			if string.format("%.2f", element) == string.format("%.2f", el) then
				return index
			end
		else
			if element == el then
				return index
			end
		end
	end
	return false
end
function RealWoodHarvester:settingsFromGui(vehicle, delimbAccuracyState, woodChipsAmountState, minCut, maxCut, cutStep, showhelpState)
	local spec = vehicle.spec_woodHarvester
	if RealWoodHarvester.showHelpIndex ~= showhelpState or spec.cutLengthMin ~= tonumber(minCut) or spec.cutLengthMax ~= tonumber(maxCut) or spec.cutLengthStep ~= tonumber(cutStep) then
		RealWoodHarvester.showHelp = RealWoodHarvester.booleans[showhelpState]
		spec.cutLengthMin = tonumber(minCut) ~= nil and tonumber(minCut) or spec.cutLengthMin
		spec.cutLengthMax = tonumber(maxCut) ~= nil and tonumber(maxCut) or spec.cutLengthMax
		spec.cutLengthStep = tonumber(cutStep) ~= nil and tonumber(cutStep) or spec.cutLengthStep
		RealWoodHarvester.showHelpIndex = showhelpState
		vehicle:saveToXmlRw(vehicle, vehicle:rwGetFullName())
	end
	if RealWoodHarvester.delimbAccuracyIndex ~= delimbAccuracyState or RealWoodHarvester.woodChipsAmountIndex ~= woodChipsAmountState then
		vehicle:rwSetGlobalConfig(RealWoodHarvester.delimbAccuracies[delimbAccuracyState], RealWoodHarvester.woodChipsAmounts[woodChipsAmountState], false)
	end
end
function RealWoodHarvester:settingsResetGui(vehicle)
	local spec = vehicle.spec_realWoodHarvester
	
	if g_currentMission.isMasterUser then
		RealWoodHarvester.gui.rwSettingGui:setDelimbAccuracy(4)
		RealWoodHarvester.gui.rwSettingGui:setWoodChipsAmount(2)
	end
	RealWoodHarvester.gui.rwSettingGui:setMinCut(spec.cutLengthMinOrig)
	RealWoodHarvester.gui.rwSettingGui:setMaxCut(spec.cutLengthMaxOrig)
	RealWoodHarvester.gui.rwSettingGui:setCutStep(spec.cutLengthStepOrig)
	RealWoodHarvester.gui.rwSettingGui:setShowhelp(1)
end
function RealWoodHarvester.prerequisitesPresent(specializations)
	return true
end
function RealWoodHarvester.registerOverwrittenFunctions(vehicleType)
end
function RealWoodHarvester.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "attachTree", 					RealWoodHarvester.attachTree)
	SpecializationUtil.registerFunction(vehicleType, "generateWoodChips", 			RealWoodHarvester.generateWoodChips)
	SpecializationUtil.registerFunction(vehicleType, "rwGetFullName", 				RealWoodHarvester.rwGetFullName)
	SpecializationUtil.registerFunction(vehicleType, "rwSetGlobalConfig", 			RealWoodHarvester.rwSetGlobalConfig)
	SpecializationUtil.registerFunction(vehicleType, "saveToXmlRw", 				RealWoodHarvester.saveToXmlRw)
	SpecializationUtil.registerFunction(vehicleType, "loadFromXmlRw", 				RealWoodHarvester.loadFromXmlRw)
end
function RealWoodHarvester:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	local spec = self.spec_realWoodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	else
		self:clearActionEventsTable(spec.event_IDs)
	end
	if self:getIsActiveForInput() or (self.getIscontrolled ~= nil and self:getIsControlled()) then
		self:loadFromXmlRw(self:rwGetFullName())
		if g_dedicatedServerInfo ~= nil then
			return
		end
		local actions = { InputAction.REAL_WOODHARVESTER_LOG_FORWARD, InputAction.REAL_WOODHARVESTER_LOG_BACKWARD, InputAction.REAL_WOODHARVESTER_CHANGE_CUT_DOWN, InputAction.REAL_WOODHARVESTER_OPEN_MENU }
		for _,actionName in pairs(actions) do
			local always = (actionName == InputAction.REAL_WOODHARVESTER_LOG_FORWARD or InputAction.REAL_WOODHARVESTER_LOG_BACKWARD) and false or false
			local _, eventID = g_inputBinding:registerActionEvent(actionName, self, RealWoodHarvester.actionCallback, true, true, always, true)
			spec.event_IDs[actionName] = eventID
			if g_inputBinding ~= nil and g_inputBinding.events ~= nil and g_inputBinding.events[eventID] ~= nil then
				if actionName == 'something with lower priority' then
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_NORMAL)
				else
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_HIGH)
				end
				g_inputBinding:setActionEventTextVisibility(eventID, RealWoodHarvester.showHelp)
			end
			local colliding = false
			_, colliding, _ = g_inputBinding:checkEventCollision(actionName)
			if colliding then
				if g_inputBinding.nameActions[actionName].bindings[1] ~= nil then
					if g_inputBinding.nameActions[actionName].bindings[1].inputString ~= nil then
						print(string.format('Warning: RealWoodHarvester got a colliding input action: %s', actionName)..' ('..g_inputBinding.nameActions[actionName].bindings[1].inputString..'). You can remap it in controls settings')
					end
				else
					print(string.format('Warning: RealWoodHarvester got a colliding input action: %s', actionName))
				end
			end
		end
	end
end
function RealWoodHarvester.registerEventListeners(vehicleType)
	for _,n in pairs( { "onLoad", "onPostLoad", "onDelete", "saveToXMLFile", "onUpdateNew", "onUpdateTickNew", "onDraw", "onReadStream", "onWriteStream", "onRegisterActionEvents", "attachTree", "rwSetGlobalConfig" } ) do
		SpecializationUtil.registerEventListener(vehicleType, n, RealWoodHarvester)
	end
end
function RealWoodHarvester:onLoad(savegame)
	self.spec_realWoodHarvester = {}
	local spec = self.spec_realWoodHarvester
	
	spec.modInitialized = false
	spec.modAllowed = true
	if self.spec_woodHarvester ~= nil then
		spec.modInitialized = true
	end
	if not spec.modInitialized then
		print("Error: RealWoodHarvester initialization failed for "..tostring(self:getFullName()).." ! ")
		return
	else
		--print("RealWoodHarvester initialized for "..tostring(self:getFullName()).." ! ")
	end
	local spec = self.spec_woodHarvester
	spec.modInitialized = true
	spec.modAllowed = true
	WoodHarvester.actionEventCutTree = Utils.overwrittenFunction(WoodHarvester.actionEventCutTree, RealWoodHarvester.actionEventCutTreeNew)
	self.woodHarvesterSplitShapeCallback = Utils.overwrittenFunction(self.woodHarvesterSplitShapeCallback, RealWoodHarvester.woodHarvesterSplitShapeCallbackNew)
	WoodHarvester.onUpdateTick = Utils.overwrittenFunction(WoodHarvester.onUpdateTick, RealWoodHarvester.onUpdateTickNew)
	WoodHarvester.onUpdate = Utils.overwrittenFunction(WoodHarvester.onUpdate, RealWoodHarvester.onUpdateNew)
	self.onDelimbTree = Utils.overwrittenFunction(self.onDelimbTree, RealWoodHarvester.onDelimbTreeNew)
	self.cutTree = Utils.overwrittenFunction(self.cutTree, RealWoodHarvester.cutTreeNew)
	
	if self.isClient then
		if RealWoodHarvester.gui == nil then
			g_gui:loadProfiles(RealWoodHarvester.modDirectory .. "gui/guiProfiles.xml")
			RealWoodHarvester.gui = {}
			RealWoodHarvester.gui["rwSettingGui"] = realWoodharvesterGui:new()
			g_gui:loadGui(RealWoodHarvester.modDirectory .. "gui/rwGui.xml", "realWoodharvesterGui", RealWoodHarvester.gui.rwSettingGui)
		end
	end
	
	spec.readyCut = 0
	spec.cutStep = 50
	spec.woodChipsCount = 0
	spec.woodChipsCutCount = 0
	spec.cutLengthMinOrig = self.spec_woodHarvester.cutLengthMin
	spec.cutLengthMaxOrig = self.spec_woodHarvester.cutLengthMax
	spec.cutLengthStepOrig = self.spec_woodHarvester.cutLengthStep
	spec.manualCut = false
	spec.input_forward = false
	spec.input_backward = false
	spec.lengthRemBottom = 0
	spec.timer = 0
	spec.timer2 = 0
	spec.moving = 0
	
	local configFile = RealWoodHarvester.confDir .. "RealWoodHarvesterConfig.xml"
	if fileExists(configFile) then
		self:loadFromXmlRw(self:rwGetFullName())
	else
		createFolder(getUserProfileAppPath().. "modsSettings/")
		createFolder(RealWoodHarvester.confDir)
		RealWoodHarvester.configXml = createXMLFile("RealWoodHarvester_XML", configFile, "RealWoodHarvesterConfig")
		self:saveToXmlRw(self)
	end
end
function RealWoodHarvester:onPostLoad(savegame)
	local spec = self.spec_realWoodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
end
function RealWoodHarvester:onDelete()
	local spec = self.spec_realWoodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
end
function RealWoodHarvester:saveToXMLFile(xmlFile, key)
	local spec = self.spec_realWoodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
end
function RealWoodHarvester:loadFromXmlRw(vehicle)
	local spec = self.spec_woodHarvester
	
	local configFile = RealWoodHarvester.confDir .. "RealWoodHarvesterConfig.xml"
	if fileExists(configFile) then
		RealWoodHarvester.configXml = loadXMLFile("RealWoodHarvester_XML", configFile)
	end
	if fileExists(configFile) then
		if getXMLBool(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.showHelp") ~= nil then
			RealWoodHarvester.showHelp = getXMLBool(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.showHelp")
		end
		if getXMLInt(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.delimbAccuracy") ~= nil then
			RealWoodHarvester.delimbAccuracy = getXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.delimbAccuracy")
		end
		if getXMLInt(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.woodChipsAmount") ~= nil then
			RealWoodHarvester.woodChipsAmount = getXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.woodChipsAmount")
		end
		if vehicle ~= nil then
			local key = "RealWoodHarvesterConfig.activeVehicles"
			local i = 0
			while true do
				local name = getXMLString(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#name")
				if name ~= nil then
					if name == vehicle then
						spec.cutLengthMin = getXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMin")
						spec.cutLengthMax = getXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMax")
						spec.cutLengthStep = getXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthStep")
						break
					end
				else
					break
				end
				i = i + 1
			end
		end
		RealWoodHarvester.delimbAccuracyIndex = getIndexFromElement(RealWoodHarvester.delimbAccuracies, RealWoodHarvester.delimbAccuracy)
		RealWoodHarvester.woodChipsAmountIndex = getIndexFromElement(RealWoodHarvester.woodChipsAmounts, RealWoodHarvester.woodChipsAmount)
		RealWoodHarvester.showHelpIndex = getIndexFromElement(RealWoodHarvester.booleans, RealWoodHarvester.showHelp)
	end
end
function RealWoodHarvester:saveToXmlRw(vehicleId, vehicle)
	local spec = vehicleId.spec_woodHarvester
	
	if RealWoodHarvester.configXml ~= nil then
		setXMLBool(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.showHelp", RealWoodHarvester.showHelp)
		setXMLInt(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.delimbAccuracy", RealWoodHarvester.delimbAccuracy)
		setXMLInt(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.woodChipsAmount", RealWoodHarvester.woodChipsAmount)
		if vehicle ~= nil then
			local key = "RealWoodHarvesterConfig.activeVehicles"
			local i = 0
			while true do
				local name = getXMLString(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#name")
				if name ~= nil then
					if name == vehicle then
						setXMLString(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMin", spec.cutLengthMin)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMax", spec.cutLengthMax)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthStep", spec.cutLengthStep)
						break
					end
				else
					setXMLString(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMin", spec.cutLengthMin)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMax", spec.cutLengthMax)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthStep", spec.cutLengthStep)
					break
				end
				i = i + 1
				local name = key .. string.format(".vehicle(%d)", i)
				if not hasXMLProperty(RealWoodHarvester.configXml, name) then
					setXMLString(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMin", spec.cutLengthMin)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthMax", spec.cutLengthMax)
						setXMLFloat(RealWoodHarvester.configXml, "RealWoodHarvesterConfig.activeVehicles"..string.format(".vehicle(%d)", i).."#cutLengthStep", spec.cutLengthStep)
					break
				end
			end
		end
		saveXMLFile(RealWoodHarvester.configXml)
	end
end
function RealWoodHarvester:onUpdateNew(onUpdate, dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local spec = self.spec_woodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if spec.attachedSplitShape == nil and spec.readyCut == 1 then
		spec.readyCut = 0
	end
	if spec.currentCutLengthLast == nil then
		spec.currentCutLengthLast = spec.currentCutLength
	end
	if spec.currentCutLength ~= spec.currentCutLengthLast then
		spec.currentCutLengthLast = spec.currentCutLength
		spec.readyCut = 0
		RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, spec.lengthRemBottom, noEventSend)
	end
	if self.isServer and spec.attachedSplitShape ~= nil then
		if spec.isAttachedSplitShapeMoving or (spec.moving and spec.timer2 < 500) then
			spec.timer = spec.timer + dt
			if spec.timer >= 50 then
				spec.timer = 0
				spec.moving = true
				local x,y,z = getWorldTranslation(spec.cutAttachReferenceNode)
				local nx,ny,nz = localDirectionToWorld(spec.cutAttachReferenceNode, 0,-1,0)
				local _, lengthRemBottom = getSplitShapePlaneExtents(spec.attachedSplitShape, x,y,z, nx,ny,nz)
				spec.lengthRemBottom = lengthRemBottom
				RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, lengthRemBottom, noEventSend)
			end
			if not spec.isAttachedSplitShapeMoving then
				if spec.moving then
					spec.timer2 = spec.timer2 + dt
					if spec.timer2 > 500 then
						spec.moving = false
						spec.timer2 = 0
					end
				end
			end
		end
	end

    -- Verify that the split shapes still exist (possible that someone has cut them)
    if self.isServer then
        local lostShape = false
        if spec.attachedSplitShape ~= nil then
            if not entityExists(spec.attachedSplitShape) then
                spec.attachedSplitShape = nil
                spec.attachedSplitShapeJointIndex = nil
                spec.isAttachedSplitShapeMoving = false
                spec.cutTimer = -1
                lostShape = true
            end
        elseif spec.curSplitShape ~= nil then
            if not entityExists(spec.curSplitShape) then
                spec.curSplitShape = nil
                lostShape = true
            end
        end
        if lostShape then
            SpecializationUtil.raiseEvent(self, "onCutTree", 0)
            if g_server ~= nil then
                g_server:broadcastEvent(WoodHarvesterOnCutTreeEvent:new(self, 0), nil, nil, self)
            end
        end
    end
    if self.isServer and (spec.attachedSplitShape ~= nil or spec.curSplitShape ~= nil) then
        if spec.cutTimer > 0 then
            if spec.cutAnimation.name ~= nil then
				if spec.woodChipsCutCount >= 6 - RealWoodHarvester.woodChipsAmount then
					spec.woodChipsCutCount = 0
					local r = math.random(-5,5) / 10
					local x,y,z = localToWorld(spec.cutNode, 0,0.7+r,0)
					self:generateWoodChips(x, y, z)
				end
                if self:getAnimationTime(spec.cutAnimation.name) > spec.cutAnimation.cutTime then
                    spec.cutTimer = 0
					spec.woodChipsCutCount = spec.woodChipsCutCount + 1
                end
            else
                spec.cutTimer = math.max(spec.cutTimer - dt, 0)
            end
        end
        local readyToCut = spec.cutTimer == 0
        -- cut
        if readyToCut then
            spec.cutTimer = -1
            local x,y,z = getWorldTranslation(spec.cutNode)
            local nx,ny,nz = localDirectionToWorld(spec.cutNode, 1,0,0)
            local yx,yy,yz = localDirectionToWorld(spec.cutNode, 0,1,0)
            local newTreeCut = false
            local currentSplitShape
            if spec.attachedSplitShapeJointIndex ~= nil then
                removeJoint(spec.attachedSplitShapeJointIndex)
                spec.attachedSplitShapeJointIndex = nil
                currentSplitShape = spec.attachedSplitShape
                spec.attachedSplitShape = nil
            else
                currentSplitShape = spec.curSplitShape
                spec.curSplitShape = nil
                newTreeCut = true
            end
            -- remember split type name for later (achievement)
            local splitTypeName = ""
            local splitType = g_splitTypeManager:getSplitTypeByIndex(getSplitType(currentSplitShape))
            if splitType ~= nil then
                splitTypeName = splitType.name
            end
            if spec.delimbOnCut then
                local xD,yD,zD = getWorldTranslation(spec.delimbNode)
                local nxD,nyD,nzD = localDirectionToWorld(spec.delimbNode, 1,0,0)
                local yxD,yyD,yzD = localDirectionToWorld(spec.delimbNode, 0,1,0)
                local vx,vy,vz = x-xD,y-yD,z-zD
                local sizeX = MathUtil.vector3Length(vx,vy,vz)
                local attachmentsRemoved = removeSplitShapeAttachments(currentSplitShape, xD+vx*0.5,yD+vy*0.5,zD+vz*0.5, nxD,nyD,nzD, yxD,yyD,yzD, sizeX*0.7+spec.delimbSizeX, spec.delimbSizeY, spec.delimbSizeZ)
				if attachmentsRemoved then
					spec.woodChipsCount = spec.woodChipsCount + (1 * RealWoodHarvester.woodChipsAmount)
				end
            end
            spec.attachedSplitShape = nil
            spec.curSplitShape = nil
            spec.prevSplitShape = currentSplitShape
            g_currentMission:removeKnownSplitShape(currentSplitShape)
            splitShape(currentSplitShape, x,y,z, nx,ny,nz, yx,yy,yz, spec.cutSizeY, spec.cutSizeZ, "woodHarvesterSplitShapeCallback", self)
            if spec.attachedSplitShape == nil then
                SpecializationUtil.raiseEvent(self, "onCutTree", 0)
                if g_server ~= nil then
                    g_server:broadcastEvent(WoodHarvesterOnCutTreeEvent:new(self, 0), nil, nil, self)
                end
            else
                if spec.delimbOnCut then
                    local xD,yD,zD = getWorldTranslation(spec.delimbNode)
                    local nxD,nyD,nzD = localDirectionToWorld(spec.delimbNode, 1,0,0)
                    local yxD,yyD,yzD = localDirectionToWorld(spec.delimbNode, 0,1,0)
                    local vx,vy,vz = x-xD,y-yD,z-zD
                    local sizeX = MathUtil.vector3Length(vx,vy,vz)
					local attachmentsRemoved = removeSplitShapeAttachments(spec.attachedSplitShape, xD+vx*3,yD+vy*3,zD+vz*3, nxD,nyD,nzD, yxD,yyD,yzD, sizeX*3+spec.delimbSizeX, spec.delimbSizeY, spec.delimbSizeZ)
					if attachmentsRemoved then
						spec.woodChipsCount = spec.woodChipsCount + (1 * RealWoodHarvester.woodChipsAmount)
					end
                end
            end
            if newTreeCut then
                local stats = g_currentMission:farmStats(self:getActiveFarm())
                -- increase tree cut counter for achievements
                stats:updateStats("cutTreeCount", 1)
                -- update the types of trees cut so far (achievement)
                if splitTypeName ~= "" then
                    stats:updateTreeTypesCut(splitTypeName)
                end
            end
        end
        -- delimb
		
		if spec.movingManual and spec.isAttachedSplitShapeMoving and not spec.input_forward and not spec.input_backward then
			spec.manualCut = true
			RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, spec.lengthRemBottom, noEventSend)
			self:onDelimbTree(false)
            if g_server ~= nil then
                g_server:broadcastEvent(WoodHarvesterOnDelimbTreeEvent:new(self, false), nil, nil, self)
            end
		end
		if spec.input_forward or spec.input_backward then
			spec.movingManual = true
		else
			spec.movingManual = false
		end
        if spec.attachedSplitShape ~= nil and spec.isAttachedSplitShapeMoving then
            if spec.delimbNode ~= nil then
				if spec.attachedSplitShapeYLast == nil then
					spec.attachedSplitShapeYLast = spec.attachedSplitShapeY
				end
				if spec.attachedSplitShapeY > spec.attachedSplitShapeYLast + 0.1 or spec.attachedSplitShapeY < spec.attachedSplitShapeYLast - 0.1 then
					spec.attachedSplitShapeYLast = spec.attachedSplitShapeY
					local random = math.random(1, 100)
					if random <= RealWoodHarvester.delimbAccuracy then
						local x,y,z = getWorldTranslation(spec.delimbNode)
						local nx,ny,nz = localDirectionToWorld(spec.delimbNode, 1,0,0)
						local yx,yy,yz = localDirectionToWorld(spec.delimbNode, 0,1,0)
						local attachmentsRemoved = removeSplitShapeAttachments(spec.attachedSplitShape, x,y,z, nx,ny,nz, yx,yy,yz, spec.delimbSizeX, spec.delimbSizeY, spec.delimbSizeZ)
						if attachmentsRemoved then
							spec.woodChipsCount = spec.woodChipsCount + (1 * RealWoodHarvester.woodChipsAmount)
						end
					end
					if spec.woodChipsCount >= 20 then
						spec.woodChipsCount = 0
						local r = math.random(-5,5) / 10
						local x,y,z = localToWorld(spec.delimbNode, 0,0.7+r,0)
						self:generateWoodChips(x, y, z)
					end
				end
            end
            if spec.cutNode ~= nil and spec.attachedSplitShapeJointIndex ~= nil then
                local x,y,z = getWorldTranslation(spec.cutAttachReferenceNode)
                local nx,ny,nz = localDirectionToWorld(spec.cutAttachReferenceNode, 0,1,0)
				local _, lengthRemTop = getSplitShapePlaneExtents(spec.attachedSplitShape, x,y,z, nx,ny,nz)
                local nx,ny,nz = localDirectionToWorld(spec.cutAttachReferenceNode, 0,-1,0)
                local _, lengthRemBottom = getSplitShapePlaneExtents(spec.attachedSplitShape, x,y,z, nx,ny,nz)
                if lengthRemTop == nil or lengthRemTop <= 0.1 or ((lengthRemBottom == nil or lengthRemBottom <= 0.2) and spec.input_backward) then
                    -- end of tree
                    removeJoint(spec.attachedSplitShapeJointIndex)
                    spec.attachedSplitShapeJointIndex = nil
                    spec.attachedSplitShape = nil
					self:onDelimbTree(false)
                    if g_server ~= nil then
                        g_server:broadcastEvent(WoodHarvesterOnDelimbTreeEvent:new(self, false), nil, nil, self)
                    end
                    SpecializationUtil.raiseEvent(self, "onCutTree", 0)
                    if g_server ~= nil then
                        g_server:broadcastEvent(WoodHarvesterOnCutTreeEvent:new(self, 0), nil, nil, self)
                    end
                else
					local canCut = false
					if spec.input_forward then
						spec.attachedSplitShapeY = spec.attachedSplitShapeY + spec.cutAttachMoveSpeed*dt
						if spec.attachedSplitShapeY >= spec.attachedSplitShapeTargetY then
							spec.attachedSplitShapeY = spec.attachedSplitShapeTargetY
							canCut = true
						end
					elseif spec.input_backward then
						spec.attachedSplitShapeY = spec.attachedSplitShapeY - spec.cutAttachMoveSpeed*dt
						if spec.attachedSplitShapeY <= spec.attachedSplitShapeTargetY then
							spec.attachedSplitShapeY = spec.attachedSplitShapeTargetY
							canCut = true
						end
					else
						spec.attachedSplitShapeY = spec.attachedSplitShapeY + spec.cutAttachMoveSpeed*dt
						if spec.attachedSplitShapeY >= spec.attachedSplitShapeTargetY then
							spec.attachedSplitShapeY = spec.attachedSplitShapeTargetY
							canCut = true
						end
					end
					if canCut then
                        self:onDelimbTree(false)
                        if g_server ~= nil then
                            g_server:broadcastEvent(WoodHarvesterOnDelimbTreeEvent:new(self, false), nil, nil, self)
                        end
                    end
                    if spec.attachedSplitShapeJointIndex ~= nil then
                        x,y,z = localToWorld(spec.cutNode, 0.3,0,0)
                        nx,ny,nz = localDirectionToWorld(spec.cutNode, 1,0,0)
                        local yx,yy,yz = localDirectionToWorld(spec.cutNode, 0,1,0)
                        local shape, minY, maxY, minZ, maxZ = findSplitShape(x,y,z, nx,ny,nz, yx,yy,yz, spec.cutSizeY, spec.cutSizeZ)
                        if shape == spec.attachedSplitShape then
                            local treeCenterX,treeCenterY,treeCenterZ = localToWorld(spec.cutNode, 0, (minY+maxY)*0.5, (minZ+maxZ)*0.5)
                            spec.attachedSplitShapeX, _, spec.attachedSplitShapeZ = worldToLocal(spec.attachedSplitShape, treeCenterX,treeCenterY,treeCenterZ)
                            self:setLastTreeDiameter((maxY-minY + maxZ-minZ)*0.5)
                        end
                        x,y,z = localToWorld(spec.attachedSplitShape, spec.attachedSplitShapeX, spec.attachedSplitShapeY, spec.attachedSplitShapeZ)
                        setJointPosition(spec.attachedSplitShapeJointIndex, 1, x,y,z)
                    end
                end
            end
        end
    end
    -- effect and sound for cut and delimb
    if self.isClient then
        -- cut
        if spec.cutAnimation.name ~= nil then
            if self:getIsAnimationPlaying(spec.cutAnimation.name) and self:getAnimationTime(spec.cutAnimation.name) < spec.cutAnimation.cutTime then
                if not spec.isCutSamplePlaying then
                    g_soundManager:playSample(spec.samples.cut)
                    spec.isCutSamplePlaying = true
                end
                g_effectManager:setFillType(spec.cutEffects, FillType.WOODCHIPS)
                g_effectManager:startEffects(spec.cutEffects)
            else
                if spec.isCutSamplePlaying then
                    g_soundManager:stopSample(spec.samples.cut)
                    spec.isCutSamplePlaying = false
                end
                g_effectManager:stopEffects(spec.cutEffects)
            end
        end
        -- delimb
        if spec.isAttachedSplitShapeMoving then
            if not spec.isDelimbSamplePlaying then
                g_soundManager:playSample(spec.samples.delimb)
                spec.isDelimbSamplePlaying = true
            end
            g_effectManager:setFillType(spec.delimbEffects, FillType.WOODCHIPS)
            g_effectManager:startEffects(spec.delimbEffects)
            g_animationManager:startAnimations(spec.forwardingNodes)
        else
            if spec.isDelimbSamplePlaying then
                g_soundManager:stopSample(spec.samples.delimb)
                spec.isDelimbSamplePlaying = false
            end
            g_effectManager:stopEffects(spec.delimbEffects)
            g_animationManager:stopAnimations(spec.forwardingNodes)
        end
		if spec.hasAttachedSplitShape then
			g_currentMission:addExtraPrintText(g_i18n:getText('REAL_WOODHARVESTER_INFO').." "..string.format('%.1f', spec.lengthRemBottom))
		end
    end
end
function RealWoodHarvester:actionEventCutTreeNew(actionEventCutTree, actionName, inputValue, callbackState, isAnalog)
	local spec = self.spec_woodHarvester
    if spec.hasAttachedSplitShape then
        if not spec.isAttachedSplitShapeMoving and self:getAnimationTime(spec.cutAnimation.name) == 1 and not spec.manualCut then
            self:cutTree(spec.currentCutLength)
		else
			self:cutTree(0)
			spec.manualCut = false
			RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, spec.lengthRemBottom, noEventSend)
			spec.readyCut = 0
        end
    elseif spec.curSplitShape ~= nil then
		if getRigidBodyType(spec.curSplitShape) == "Dynamic" or getRigidBodyType(spec.curSplitShape) == "Kinematic" then
			self:attachTree(false)
		elseif getRigidBodyType(spec.curSplitShape) == "Static" then
			self:cutTree(0)
		end
    end
end
function RealWoodHarvester:cutTreeNew(cutTree, length, noEventSend)
    local spec = self.spec_woodHarvester
	
    WoodHarvesterCutTreeEvent.sendEvent(self, length, noEventSend)
    if self.isServer then
        if length == 0 then
            if spec.attachedSplitShape ~= nil or spec.curSplitShape ~= nil then
                spec.cutTimer = 100
                if spec.cutAnimation.name ~= nil then
                    self:setAnimationTime(spec.cutAnimation.name, 0, true)
                    self:playAnimation(spec.cutAnimation.name, spec.cutAnimation.speedScale, self:getAnimationTime(spec.cutAnimation.name))
                end
            end
        elseif length ~= 0 and spec.attachedSplitShape ~= nil then
			spec.attachedSplitShapeTargetY = spec.attachedSplitShapeLastCutY + length
            self:onDelimbTree(true)
            if g_server ~= nil then
                g_server:broadcastEvent(WoodHarvesterOnDelimbTreeEvent:new(self, true), nil, nil, self)
            end
        end
    end
end
function RealWoodHarvester:attachTree(noEventSend)
	local spec = self.spec_woodHarvester
	
	WoodHarvesterAttachTreeEvent.sendEvent(self, noEventSend)
	if self.isServer then
		local x,y,z = getWorldTranslation(spec.cutNode)
		local nx,ny,nz = localDirectionToWorld(spec.cutNode, 1,0,0)
		local yx,yy,yz = localDirectionToWorld(spec.cutNode, 0,1,0)
		local minY,maxY, minZ,maxZ = testSplitShape(spec.curSplitShape, x,y,z, nx,ny,nz, yx,yy,yz, spec.cutSizeY, spec.cutSizeZ)
		if minY ~= nil and minY ~= 0 then
			self:woodHarvesterSplitShapeCallback(spec.curSplitShape, false, true, minY, maxY, minZ, maxZ)
		end
	end
	self:setAnimationTime(spec.cutAnimation.name, 1, true)
	spec.curSplitShape = nil
end
function RealWoodHarvester:onDelimbTreeNew(onDelimbTree, state)
	local spec = self.spec_woodHarvester
	
	self.spec_woodHarvester.isAttachedSplitShapeMoving = state
	if not state and not spec.manualCut then
		if spec.readyCut == 1 then
			self:cutTree(0)
			spec.readyCut = 0
		else
			spec.readyCut = 1
		end
	end
end
function RealWoodHarvester:woodHarvesterSplitShapeCallbackNew(woodHarvesterSplitShapeCallback, shape, isBelow, isAbove, minY, maxY, minZ, maxZ)
    local spec = self.spec_woodHarvester
	
    g_currentMission:addKnownSplitShape(shape)
    if spec.attachedSplitShape == nil and isAbove and not isBelow and spec.cutAttachNode ~= nil and spec.cutAttachReferenceNode ~= nil then
        spec.attachedSplitShape = shape
        -- Current tree center (mid of cut area)
        local treeCenterX,treeCenterY,treeCenterZ = localToWorld(spec.cutNode, 0, (minY+maxY)*0.5, (minZ+maxZ)*0.5)
        -- Target tree center (half tree size in front of the reference node)
        local x,y,z = localToWorld(spec.cutAttachReferenceNode, 0, 0, (maxZ-minZ)*0.5)
        local dx,dy,dz = localDirectionToWorld(shape, 0,0,1)
        local upx,upy,upz = localDirectionToWorld(spec.cutAttachReferenceNode, 0,1,0)
        local sideX,sideY,sizeZ = MathUtil.crossProduct(upx,upy,upz, dx,dy,dz)
        dx,dy,dz = MathUtil.crossProduct(sideX,sideY,sizeZ, upx,upy,upz) -- Note: we want the up axis to be exact, thus orthogonalize the direction here
        I3DUtil.setWorldDirection(spec.cutAttachHelperNode, dx,dy,dz, upx,upy,upz, 2)
        local constr = JointConstructor:new()
        constr:setActors(spec.cutAttachNode, shape)
        -- Note: we assume that the direction of the tree is equal to the y axis
        constr:setJointTransforms(spec.cutAttachHelperNode, shape)
        constr:setJointWorldPositions(x,y,z, treeCenterX,treeCenterY,treeCenterZ)
        constr:setRotationLimit(0, 0, 0)
        constr:setRotationLimit(1, 0, 0)
        constr:setRotationLimit(2, 0, 0)
        constr:setEnableCollision(false)
        spec.attachedSplitShapeJointIndex = constr:finalize()
        if spec.cutReleasedComponentJoint ~= nil then
            spec.cutReleasedComponentJointRotLimitX = math.pi*0.9
            if spec.cutReleasedComponentJoint.jointIndex ~= 0 then
                setJointRotationLimit(spec.cutReleasedComponentJoint.jointIndex, 0, true, 0, spec.cutReleasedComponentJointRotLimitX)
            end
        end
        if spec.cutReleasedComponentJoint2 ~= nil then
            spec.cutReleasedComponentJoint2RotLimitX = math.pi*0.9
            if spec.cutReleasedComponentJoint2.jointIndex ~= 0 then
                setJointRotationLimit(spec.cutReleasedComponentJoint2.jointIndex, 0, true, -spec.cutReleasedComponentJoint2RotLimitX, spec.cutReleasedComponentJoint2RotLimitX)
            end
        end
        spec.attachedSplitShapeX, spec.attachedSplitShapeY, spec.attachedSplitShapeZ = worldToLocal(shape, treeCenterX,treeCenterY,treeCenterZ)
        spec.attachedSplitShapeLastCutY = spec.attachedSplitShapeY
        spec.attachedSplitShapeStartY = spec.attachedSplitShapeY
        spec.attachedSplitShapeTargetY = spec.attachedSplitShapeY
        local radius = ((maxY - minY) + (maxZ - minZ)) / 4
        SpecializationUtil.raiseEvent(self, "onCutTree", radius)
        if g_server ~= nil then
            g_server:broadcastEvent(WoodHarvesterOnCutTreeEvent:new(self, radius), nil, nil, self)
        end
		local x,y,z = getWorldTranslation(spec.cutAttachReferenceNode)
		local nx,ny,nz = localDirectionToWorld(spec.cutAttachReferenceNode, 0,-1,0)
		local _, lengthRemBottom = getSplitShapePlaneExtents(spec.attachedSplitShape, x,y,z, nx,ny,nz)
		spec.lengthRemBottom = lengthRemBottom
		RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, lengthRemBottom, noEventSend)
    end
end
function RealWoodHarvester:onUpdateTickNew(onUpdateTick, dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local spec = self.spec_woodHarvester
	
    spec.warnInvalidTree = false
    spec.warnInvalidTreeRadius = false
    spec.warnTreeNotOwned = false
    if self:getIsTurnedOn() then
        if spec.attachedSplitShape == nil and spec.cutNode ~= nil then
            local x,y,z = getWorldTranslation(spec.cutNode)
            local nx,ny,nz = localDirectionToWorld(spec.cutNode, 1,0,0)
            local yx,yy,yz = localDirectionToWorld(spec.cutNode, 0,1,0)
            if spec.curSplitShape == nil and (spec.cutReleasedComponentJoint == nil or spec.cutReleasedComponentJointRotLimitX == 0) then
                local shape, minY, maxY, minZ, maxZ = findSplitShape(x,y,z, nx,ny,nz, yx,yy,yz, spec.cutSizeY, spec.cutSizeZ)
                if shape ~= 0 then
                    local splitType = g_splitTypeManager:getSplitTypeByIndex(getSplitType(shape))
                    if splitType == nil or not splitType.allowsWoodHarvester then
                        spec.warnInvalidTree = true
                    else
                        if g_currentMission.accessHandler:canFarmAccessLand(self:getActiveFarm(), x, z) then
                            local treeDx,treeDy,treeDz = localDirectionToWorld(shape, 0,1,0) -- wood harvester trees always grow in the y direction
                            local cosTreeAngle = MathUtil.dotProduct(nx,ny,nz, treeDx,treeDy,treeDz)
                            -- Only allow cutting if the cut header is approximately parallel to the tree (70 offset)
                            if cosTreeAngle >= 0.35 then
                                local radius = math.max(maxY-minY, maxZ-minZ)*0.5 * cosTreeAngle
                                if radius > spec.cutMaxRadius then
                                    spec.warnInvalidTreeRadius = true
                                else
                                    self:setLastTreeDiameter(math.max(maxY-minY, maxZ-minZ))
                                    spec.curSplitShape = shape
                                end
                            end
                        else
                            spec.warnTreeNotOwned = true
                        end
                    end
                end
            end
            if spec.curSplitShape ~= nil then
                local minY,maxY, minZ,maxZ = testSplitShape(spec.curSplitShape, x,y,z, nx,ny,nz, yx,yy,yz, spec.cutSizeY, spec.cutSizeZ)
                if minY == nil then
                    spec.curSplitShape = nil
                else
                    -- check if cut would be below y=0 (tree CoSy)
                    local cutTooLow = false
                    _,y,_ = localToLocal(spec.cutNode, spec.curSplitShape, 0,minY,minZ)
                    cutTooLow = cutTooLow or y < 0.01
                    _,y,_ = localToLocal(spec.cutNode, spec.curSplitShape, 0,minY,maxZ)
                    cutTooLow = cutTooLow or y < 0.01
                    _,y,_ = localToLocal(spec.cutNode, spec.curSplitShape, 0,maxY,minZ)
                    cutTooLow = cutTooLow or y < 0.01
                    _,y,_ = localToLocal(spec.cutNode, spec.curSplitShape, 0,maxY,maxZ)
                    cutTooLow = cutTooLow or y < 0.01
                    if cutTooLow then
                        spec.curSplitShape = nil
                    end
                end
            end
            if spec.curSplitShape == nil and spec.cutTimer > -1 then
                SpecializationUtil.raiseEvent(self, "onCutTree", 0)
                if g_server ~= nil then
                    g_server:broadcastEvent(WoodHarvesterOnCutTreeEvent:new(self, 0), nil, nil, self)
                end
            end
        end
    end
    if self.isServer then
        if spec.attachedSplitShape == nil then
            if spec.cutReleasedComponentJoint ~= nil and spec.cutReleasedComponentJointRotLimitX ~= 0 then
                spec.cutReleasedComponentJointRotLimitX = math.max(0, spec.cutReleasedComponentJointRotLimitX - spec.cutReleasedComponentJointRotLimitXSpeed*dt)
                setJointRotationLimit(spec.cutReleasedComponentJoint.jointIndex, 0, true, 0, spec.cutReleasedComponentJointRotLimitX)
            end
            if spec.cutReleasedComponentJoint2 ~= nil and spec.cutReleasedComponentJoint2RotLimitX ~= 0 then
                spec.cutReleasedComponentJoint2RotLimitX = math.max(spec.cutReleasedComponentJoint2RotLimitX-spec.cutReleasedComponentJoint2RotLimitXSpeed*dt, 0)
                setJointRotationLimit(spec.cutReleasedComponentJoint2.jointIndex, 0, true, -spec.cutReleasedComponentJoint2RotLimitX, spec.cutReleasedComponentJoint2RotLimitX)
            end
        end
    end
    if self.isServer then
        if self.playDelayedGrabAnimationTime ~= nil then
            if self.playDelayedGrabAnimationTime < g_currentMission.time then
                self.playDelayedGrabAnimationTime = nil
                if self:getAnimationTime(spec.grabAnimation.name) > 0 then
                    if spec.grabAnimation.name ~= nil and spec.attachedSplitShape == nil then
                        if spec.grabAnimation.speedScale > 0 then
                            self:setAnimationStopTime(spec.grabAnimation.name, 0)
                        else
                            self:setAnimationStopTime(spec.grabAnimation.name, 1)
                        end
                        self:playAnimation(spec.grabAnimation.name, -spec.grabAnimation.speedScale, self:getAnimationTime(spec.grabAnimation.name), false)
                    end
                end
            end
        end
    end
    if self.isClient then
        local actionEvent = spec.actionEvents[InputAction.IMPLEMENT_EXTRA2]
        if actionEvent ~= nil then
            local showAction = false
            if spec.hasAttachedSplitShape then
                if not spec.isAttachedSplitShapeMoving and self:getAnimationTime(spec.cutAnimation.name) == 1 then
                    showAction = true
                end
				g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText("action_woodHarvesterCut"))
            elseif spec.curSplitShape ~= nil then
                showAction = true
				if getRigidBodyType(spec.curSplitShape) == "Dynamic" or getRigidBodyType(spec.curSplitShape) == "Kinematic" then
					g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText("action_woodHarvesterGrab"))
				elseif getRigidBodyType(spec.curSplitShape) == "Static" then
					g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText("action_woodHarvesterCut"))
				end
            end
            g_inputBinding:setActionEventActive(actionEvent.actionEventId, showAction)
			
        end
        actionEvent = spec.actionEvents[InputAction.IMPLEMENT_EXTRA3]
        if actionEvent ~= nil then
            g_inputBinding:setActionEventActive(actionEvent.actionEventId, not spec.isAttachedSplitShapeMoving)
            if not spec.isAttachedSplitShapeMoving then
                g_inputBinding:setActionEventText(actionEvent.actionEventId, string.format(g_i18n:getText("action_woodHarvesterChangeCutLength"), string.format("%.1f", spec.currentCutLength)))
				g_inputBinding:setActionEventTextPriority(actionEvent.actionEventId, GS_PRIO_HIGH)
            end
        end
		local spec2 = self.spec_realWoodHarvester
		if self:getIsActiveForInput() or (self.getIscontrolled ~= nil and self:getIsControlled()) then
			if spec2.event_IDs ~= nil and g_dedicatedServerInfo == nil then
				for actionName,eventID in pairs(spec2.event_IDs) do
					g_inputBinding:setActionEventTextVisibility(eventID, RealWoodHarvester.showHelp)
					if actionName == InputAction.REAL_WOODHARVESTER_LOG_FORWARD then
						g_inputBinding:setActionEventActive(eventID, spec.hasAttachedSplitShape)
						g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_REAL_WOODHARVESTER_LOG_FORWARD'))
					elseif actionName == InputAction.REAL_WOODHARVESTER_LOG_BACKWARD then
						g_inputBinding:setActionEventActive(eventID, spec.hasAttachedSplitShape)
						g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_REAL_WOODHARVESTER_LOG_BACKWARD'))
					elseif actionName == InputAction.REAL_WOODHARVESTER_CHANGE_CUT_DOWN then
						g_inputBinding:setActionEventActive(eventID, not spec.isAttachedSplitShapeMoving)
						g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_REAL_WOODHARVESTER_CHANGE_CUT_DOWN'))
					end
				end
			end
		end
    end
end
function RealWoodHarvester:actionCallback(actionName, keyStatus, arg4, arg5, arg6)
	local spec = self.spec_woodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if self:getIsActive() then
		if keyStatus > 0 then
			if actionName == 'REAL_WOODHARVESTER_LOG_FORWARD' then
				spec.input_forward = true
				spec.input_backward = false
				RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, spec.lengthRemBottom, noEventSend)
				self:cutTree(spec.cutStep, false)
			elseif actionName == 'REAL_WOODHARVESTER_LOG_BACKWARD' then
				spec.input_backward = true
				spec.input_forward = false
				RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, spec.lengthRemBottom, noEventSend)
				self:cutTree(-spec.cutStep, false)
			elseif actionName == 'REAL_WOODHARVESTER_CHANGE_CUT_DOWN' then
				if not spec.isAttachedSplitShapeMoving then
					spec.currentCutLength = spec.currentCutLength - spec.cutLengthStep
					if spec.currentCutLength < spec.cutLengthMin - 0.0001 then
						spec.currentCutLength = spec.cutLengthMax
					end
				end
			elseif actionName == 'REAL_WOODHARVESTER_OPEN_MENU' then
				if RealWoodHarvester.gui.rwSettingGui.isOpen then
					RealWoodHarvester.gui.rwSettingGui:onClickBack()
				elseif g_gui.currentGui == nil then
					RealWoodHarvester.gui.rwSettingGui:setTarget(self)
					RealWoodHarvester.gui.rwSettingGui:setDelimbAccuracy(RealWoodHarvester.delimbAccuracyIndex)
					RealWoodHarvester.gui.rwSettingGui:setWoodChipsAmount(RealWoodHarvester.woodChipsAmountIndex)
					RealWoodHarvester.gui.rwSettingGui:setMinCut(spec.cutLengthMin)
					RealWoodHarvester.gui.rwSettingGui:setMaxCut(spec.cutLengthMax)
					RealWoodHarvester.gui.rwSettingGui:setCutStep(spec.cutLengthStep)
					RealWoodHarvester.gui.rwSettingGui:setShowhelp(RealWoodHarvester.showHelpIndex)
					
					RealWoodHarvester.gui.rwSettingGui.delimbAccuracy:setDisabled(not g_currentMission.isMasterUser)
					RealWoodHarvester.gui.rwSettingGui.woodChipsAmount:setDisabled(not g_currentMission.isMasterUser)
					g_gui:showGui("realWoodharvesterGui")
				end
			end
		elseif keyStatus == 0 then
			if actionName == 'REAL_WOODHARVESTER_LOG_FORWARD' then
				spec.input_forward = false
				RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, spec.lengthRemBottom, noEventSend)
			elseif actionName == 'REAL_WOODHARVESTER_LOG_BACKWARD' then
				spec.input_backward = false
				RealWoodHarvesterMoveEvent.sendEvent(self, spec.input_forward, spec.input_backward, spec.manualCut, spec.readyCut, spec.lengthRemBottom, noEventSend)
			end
		end
	end
end
function RealWoodHarvester:generateWoodChips(x, y, z)
	local litersToDrop = math.max(g_densityMapHeightManager:getMinValidLiterValue(FillType.WOODCHIPS), 15 * 1)
	local ex = x + 0
    local ey = y - 0.1
    local ez = z + 0
	
	local outerRadius = DensityMapHeightUtil.getDefaultMaxRadius(FillType.WOODCHIPS)
	local dropped, lineOffset = DensityMapHeightUtil.tipToGroundAroundLine(self, litersToDrop, FillType.WOODCHIPS, x,y,z, ex,ey,ez, 0, outerRadius, 0, true, nil, true)
end
function RealWoodHarvester:onDraw()
	local spec = self.spec_realWoodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
end
function RealWoodHarvester:rwGetFullName()
    local name = self:getName()
    local storeItem = g_storeManager:getItemByXMLFilename(self.configFileName)
    if storeItem ~= nil then
        local brand = g_brandManager:getBrandByIndex(storeItem.brandIndex)
        if brand ~= nil then
            name = brand.title .. " " .. name
        end
    end
    return name
end
function RealWoodHarvester:rwSetGlobalConfig(delimbAccuracy, woodChipsAmount, noEventSend)
	local spec = self.spec_woodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	RealWoodHarvester.delimbAccuracy = delimbAccuracy
	RealWoodHarvester.woodChipsAmount = woodChipsAmount
	
	RealWoodHarvester.delimbAccuracyIndex = getIndexFromElement(RealWoodHarvester.delimbAccuracies, RealWoodHarvester.delimbAccuracy)
	RealWoodHarvester.woodChipsAmountIndex = getIndexFromElement(RealWoodHarvester.woodChipsAmounts, RealWoodHarvester.woodChipsAmount)
	
	self:saveToXmlRw(self)
	RealWoodHarvesterSetGlobalConfigEvent.sendEvent(self, delimbAccuracy, woodChipsAmount, noEventSend)
end
function RealWoodHarvester:onReadStream(streamId, connection)
	local spec = self.spec_woodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	local delimbAccuracy = streamReadInt32(streamId)
	local woodChipsAmount = streamReadInt32(streamId)
	if delimbAccuracy ~= nil and woodChipsAmount ~= nil then
		self:rwSetGlobalConfig(delimbAccuracy, woodChipsAmount, true)
	end
	local input_forward = streamReadBool(streamId)
	local input_backward = streamReadBool(streamId)
	local manualCut = streamReadBool(streamId)
	local readyCut = streamReadInt16(streamId)
	local lengthRemBottom = streamReadFloat32(streamId)
	if input_forward ~= nil and input_backward ~= nil and manualCut ~= nil and readyCut ~= nil and lengthRemBottom ~= nil then
		spec.input_forward = input_forward
		spec.input_backward = input_backward
		spec.manualCut = manualCut
		spec.readyCut = readyCut
		spec.lengthRemBottom = lengthRemBottom
	end
end
function RealWoodHarvester:onWriteStream(streamId, connection)
	local spec = self.spec_woodHarvester
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	streamWriteInt32(streamId, RealWoodHarvester.delimbAccuracy)
	streamWriteInt32(streamId, RealWoodHarvester.woodChipsAmount)
	streamWriteBool(streamId, spec.input_forward)
	streamWriteBool(streamId, spec.input_backward)
	streamWriteBool(streamId, spec.manualCut)
	streamWriteInt16(streamId, spec.readyCut)
	streamWriteFloat32(streamId, spec.lengthRemBottom)
end

RealWoodHarvesterSetGlobalConfigEvent = {}
RealWoodHarvesterSetGlobalConfigEvent_mt = Class(RealWoodHarvesterSetGlobalConfigEvent, Event)
InitEventClass(RealWoodHarvesterSetGlobalConfigEvent, "RealWoodHarvesterSetGlobalConfigEvent")
function RealWoodHarvesterSetGlobalConfigEvent:emptyNew()
    local self = Event:new(RealWoodHarvesterSetGlobalConfigEvent_mt)
    self.className="RealWoodHarvesterSetGlobalConfigEvent"
    return self
end
function RealWoodHarvesterSetGlobalConfigEvent:new(object, delimbAccuracy, woodChipsAmount)
	local self = RealWoodHarvesterSetGlobalConfigEvent:emptyNew()
	self.object = object
	self.delimbAccuracy = delimbAccuracy
	self.woodChipsAmount = woodChipsAmount
	return self
end
function RealWoodHarvesterSetGlobalConfigEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.delimbAccuracy = streamReadInt32(streamId)
	self.woodChipsAmount = streamReadInt32(streamId)
    self:run(connection)
end
function RealWoodHarvesterSetGlobalConfigEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteInt32(streamId, self.delimbAccuracy)
	streamWriteInt32(streamId, self.woodChipsAmount)
end
function RealWoodHarvesterSetGlobalConfigEvent:run(connection)
	if self.object ~= nil then
		self.object:rwSetGlobalConfig(self.delimbAccuracy, self.woodChipsAmount, true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(RealWoodHarvesterSetGlobalConfigEvent:new(self.object, self.delimbAccuracy, self.woodChipsAmount), nil, connection, self.object)
	end
end
function RealWoodHarvesterSetGlobalConfigEvent.sendEvent(vehicle, delimbAccuracy, woodChipsAmount, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(RealWoodHarvesterSetGlobalConfigEvent:new(vehicle, delimbAccuracy, woodChipsAmount), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(RealWoodHarvesterSetGlobalConfigEvent:new(vehicle, delimbAccuracy, woodChipsAmount))
		end
	end
end

WoodHarvesterAttachTreeEvent = {}
WoodHarvesterAttachTreeEvent_mt = Class(WoodHarvesterAttachTreeEvent, Event)
InitEventClass(WoodHarvesterAttachTreeEvent, "WoodHarvesterAttachTreeEvent")
function WoodHarvesterAttachTreeEvent:emptyNew()
    local self = Event:new(WoodHarvesterAttachTreeEvent_mt)
    self.className="WoodHarvesterAttachTreeEvent"
    return self
end
function WoodHarvesterAttachTreeEvent:new(object)
	local self = WoodHarvesterAttachTreeEvent:emptyNew()
	self.object = object
	return self
end
function WoodHarvesterAttachTreeEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
    self:run(connection)
end
function WoodHarvesterAttachTreeEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
end
function WoodHarvesterAttachTreeEvent:run(connection)
	if self.object ~= nil then
		self.object:attachTree(true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(WoodHarvesterAttachTreeEvent:new(self.object), nil, connection, self.object)
	end
end
function WoodHarvesterAttachTreeEvent.sendEvent(vehicle, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(WoodHarvesterAttachTreeEvent:new(vehicle), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(WoodHarvesterAttachTreeEvent:new(vehicle))
		end
	end
end

RealWoodHarvesterMoveEvent = {}
RealWoodHarvesterMoveEvent_mt = Class(RealWoodHarvesterMoveEvent, Event)
InitEventClass(RealWoodHarvesterMoveEvent, "RealWoodHarvesterMoveEvent")
function RealWoodHarvesterMoveEvent:emptyNew()
    local self = Event:new(RealWoodHarvesterMoveEvent_mt)
    self.className="RealWoodHarvesterMoveEvent"
    return self
end
function RealWoodHarvesterMoveEvent:new(object, forward, backward, manualCut, readyCut, lengthRemBottom)
	local self = RealWoodHarvesterMoveEvent:emptyNew()
	self.object = object
	self.forward = forward
	self.backward = backward
	self.manualCut = manualCut
	self.readyCut = readyCut
	self.lengthRemBottom = lengthRemBottom
	return self
end
function RealWoodHarvesterMoveEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.forward = streamReadBool(streamId)
	self.backward = streamReadBool(streamId)
	self.manualCut = streamReadBool(streamId)
	self.readyCut = streamReadInt16(streamId)
	self.lengthRemBottom = streamReadFloat32(streamId)
    self:run(connection)
end
function RealWoodHarvesterMoveEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.forward)
	streamWriteBool(streamId, self.backward)
	streamWriteBool(streamId, self.manualCut)
	streamWriteInt16(streamId, self.readyCut)
	streamWriteFloat32(streamId, self.lengthRemBottom)
end
function RealWoodHarvesterMoveEvent:run(connection)
	if self.object ~= nil then
		self.object.spec_woodHarvester.input_forward = self.forward
		self.object.spec_woodHarvester.input_backward = self.backward
		self.object.spec_woodHarvester.manualCut = self.manualCut
		self.object.readyCut = self.readyCut
		self.object.spec_woodHarvester.lengthRemBottom = self.lengthRemBottom
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(RealWoodHarvesterMoveEvent:new(self.object, self.forward, self.backward, self.manualCut, self.readyCut, self.lengthRemBottom), nil, connection, self.object)
	end
end
function RealWoodHarvesterMoveEvent.sendEvent(vehicle, forward, backward, manualCut, readyCut, lengthRemBottom, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(RealWoodHarvesterMoveEvent:new(vehicle, forward, backward, manualCut, readyCut, lengthRemBottom), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(RealWoodHarvesterMoveEvent:new(vehicle, forward, backward, manualCut, readyCut, lengthRemBottom))
		end
	end
end